/*------------------------------------------------------------------------------
 *       Filename:  prompter.cpp
 *          Usage:  prompter -learn <config_file>
 *                  prompter -monitor <config_file>
 *    Description:  Prompter program for learning activity patterns and
 *                  monitoring for prompt events.
 *        Version:  1.1
 *        Created:  08/01/2012
 *       Compiler:  gcc
 *        Authors:  Larry Holder, holder@eecs.wsu.edu
 *                  Narayanan C Krishnan, ckn@eecs.wsu.edu
 *                  Diane J. Cook, cook@eecs.wsu.edu
 *   Organization:  Washington State University
 * ---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <sstream>
#include <string>
#include <map>
#include "casas.h"
#include "connection.h"
#include "prompter.h"

using namespace std;

int main (int argc, char** argv)
{
	int features[NUM_FEATURES];

	// Detect prompter mode
	if (argc > 1)
	{
		if (strcmp (argv[1], "-learn") == 0)
		{
			prompterMode = LEARN_MODE;
		} else if (strcmp (argv[1], "-monitor") == 0)
		{
			prompterMode = MONITOR_MODE;
		} else if (strcmp (argv[1], "-generate") == 0)
		{
			prompterMode = GENERATE_MODE;
		} else if (strcmp (argv[1], "-test") == 0)
		{
			prompterMode = TEST_MODE;
		} else {
			printf ("Error: Unknown mode\n");
			PrompterUsage();
			exit(1);
		}
	} else {
		printf ("Error: No mode specified\n");
		PrompterUsage();
		exit(1);
	}

	// Read configuration file
	if (argc > 2)
	{
		ReadConfig (argv[2]);
	} else {
		printf ("Error: No configuration file specified\n");
		PrompterUsage();
		exit(1);
	}
	AddPrompterActivities();

	if (OutputLevel > 0)
		PrintParms();

	AllocateMemory();
	if (prompterMode == LEARN_MODE)
	{
		ReadData();
		if (NumData == 0)
		{
			printf ("No sensor data events available\n");
			exit(1);
		}
		if (OutputLevel > 0)
		{
			PrintActivityInstances (0, NumActivities - 1);
		}
		AddPrompterEvents();
		if (OutputLevel > 0)
		{
			PrintActivityInstances (NumActivities, NumActivities + NUM_PROMPTER_ACTIVITIES - 1);
		}
		LearnActivityPatterns();
		WriteActivityPatterns();
	}
	if (prompterMode == MONITOR_MODE)
	{
		ReadActivityPatterns();
		CheckPromptActivityPatterns();
		// Initialize and connect to CASAS middleware
		InitMonitorInterface();
		cbot->connect();
	}
	if (prompterMode == GENERATE_MODE)
	{
		if (argc < 7)
		{
			printf ("Error: start and end time stamps not all provided in -generate mode\n");
			PrompterUsage();
			exit(1);
		}
		SetStartTime (argv[3], argv[4], features);
		SetEndTime (argv[5], argv[6], features);
		ReadActivityPatterns();
		GenerateData (GetStartTime (features), GetEndTime (features));
	}
	if (prompterMode == TEST_MODE)
	{
		ReadData();
		if (NumData == 0)
		{
			printf ("No sensor data events available\n");
			exit(1);
		}
		if (OutputLevel > 0)
		{
			PrintActivityInstances (0, NumActivities - 1);
		}
		AddPrompterEvents();
		if (OutputLevel > 0)
		{
			PrintActivityInstances (NumActivities, NumActivities + NUM_PROMPTER_ACTIVITIES - 1);
		}
		ReadActivityPatterns();
		TestData();
	}
	FreeMemory();
	return 0;
}


// Output warning if a prompt activity has no pattern
void CheckPromptActivityPatterns ()
{
	int i, activityNum;

	for (i = 0; i < NumPromptActivities; i++)
	{
		activityNum = FindActivity (promptactivitynames[i]);
		if (NumPatternsPerActivity[activityNum] == 0)
		{
			printf ("Warning: No pattern for prompt activity: %s. Prompting deactivated for this activity.\n",
					promptactivitynames[i]);
		}
	}
}


// Adds prompter activity names to the activitynames array
void AddPrompterActivities ()
{
	int i;

	activitynames = (char **) realloc (activitynames, sizeof(char *) * (NumActivities + NUM_PROMPTER_ACTIVITIES));
	for (i = 0; i < NUM_PROMPTER_ACTIVITIES; i++)
	{
		activitynames[NumActivities + i] = (char *) malloc (sizeof(char) * MAXSTR);
		strcpy (activitynames[NumActivities + i], prompterActivityNames[i]);
	}
}


/* Read each line of the config file and set program parameters accordingly. */
void ReadConfig (char *configFileName)
{
    FILE *fp;
    char *cptr, buffer[MAX_LINE_LEN], *tok;

    fp = fopen(configFileName, "r");
    if (fp == NULL)
    {
        printf("Unable to open the configuration file: %s\n", configFileName);
        exit(1);
    }

    cptr =  fgets (buffer, MAX_LINE_LEN, fp);

    while (cptr != NULL)                // Process each line of the config file
    {
        tok = strtok(cptr, " \n");		// Process each string separated by " "
        if (strcmp (tok, "data") == 0)	// Data file name(s)
        {
            tok = strtok(NULL, " \n");
            NumFiles = 0;
            if (tok == NULL)
            {
                printf ("No data file specified\n");
                exit(1);
            }
            while (tok != NULL)
            {
                if (NumFiles == 0)
                    filenames = (char **) malloc(sizeof(char *));
                else filenames =
                    (char **) realloc(filenames, sizeof(char *) * (NumFiles + 1));
                filenames[NumFiles] = (char *) malloc(sizeof(char ) * MAXSTR);
                strcpy(filenames[NumFiles], tok);
                NumFiles++;
                tok = strtok(NULL, " \n");
            }
        }
        else if (strcmp (tok, "pattern") == 0)	// Specify file name to read/write patterns
        {
            tok = strtok(NULL, " \n");
            if (tok != NULL)
            {
                patternfname = (char *) malloc(sizeof(char) * MAXSTR);
                strcpy (patternfname, tok);
            }
        }
        else if (strcmp (tok, "sensor") == 0)  // List of sensor names
        {
            tok = strtok (NULL, " \n");
            while (tok != NULL)
            {
                if (NumSensors == 0)
                    sensornames = (char **) malloc(sizeof(char *));
                else sensornames =
                    (char **) realloc(sensornames, sizeof(char *) * (NumSensors+1));
                sensornames[NumSensors] = (char *) malloc(sizeof(char) * MAXSTR);
                strcpy(sensornames[NumSensors], tok);
                NumSensors++;
                tok = strtok(NULL, " \n");
            }
        }

        else if (strcmp (tok, "activity") == 0)  // List of activity labels
        {
            tok = strtok (NULL, " \n");
            while (tok != NULL)
            {
                if (NumActivities == 0)
                    activitynames = (char **) malloc(sizeof(char *));
                else activitynames = (char **) realloc (activitynames, sizeof(char *) * (NumActivities+1));
                activitynames[NumActivities] = (char *) malloc (sizeof(char) * MAXSTR);
                strcpy (activitynames[NumActivities], tok);
                NumActivities++;
                tok = strtok (NULL, " \n");
            }
        }
        else if (strcmp (tok, "prompt") == 0)  // List of prompt activity labels
        {
            tok = strtok (NULL, " \n");
            while (tok != NULL)
            {
                if (NumPromptActivities == 0)
                    promptactivitynames = (char **) malloc (sizeof(char *));
                else promptactivitynames = (char **) realloc (promptactivitynames, sizeof(char *) * (NumPromptActivities+1));
                promptactivitynames[NumPromptActivities] = (char *) malloc (sizeof(char) * MAXSTR);
                strcpy (promptactivitynames[NumPromptActivities], tok);
                NumPromptActivities++;
                tok = strtok (NULL, " \n");
            }
        } else {
        	printf ("Ignoring %s line in configuration file\n", tok);
        }
        cptr =  fgets (buffer, MAX_LINE_LEN, fp);  // Read next line
    }
    fclose (fp);

    // Check configuration
    if (NumActivities == 0)
    {
    	printf ("No activities defined in configuration file.\n");
    	exit(1);
    }
    if ((prompterMode == MONITOR_MODE) && (NumPromptActivities == 0))
    {
    	printf ("No prompt activities defined in configuration file.\n");
    	exit(1);
    }
    if (patternfname == NULL)
    {
    	printf ("No pattern file specified in configuration file.\n");
    	exit(1);
    }
    if ((prompterMode != MONITOR_MODE) && (NumFiles == 0))
    {
    	printf ("No data file in configuration file.\n");
    	exit(1);
    }
}

/* Print program parameter values */
void PrintParms()
{
    int i;

    printf("*****************************\n");
    printf("**** Prompter Parameters ****\n\n");

    printf ("Mode: ");
    switch (prompterMode) {
    case LEARN_MODE:
    	printf ("learn\n\n");
    	break;
    case MONITOR_MODE:
    	printf ("monitor\n\n");
    	break;
    case GENERATE_MODE:
    	printf ("generate\n\n");
    }

    printf("Pattern file: %s\n\n", patternfname);

    printf("Data files (%d):", NumFiles);
    for (i=0; i<NumFiles; i++)
    	printf(" %s", filenames[i]);
    printf("\n\n");

    printf("Sensor names (%d):", NumSensors);
    for (i=0; i<NumSensors; i++)
        printf(" %s", sensornames[i]);
    printf("\n\n");

    printf("Activity labels (%d):", NumActivities);
    for (i=0; i<NumActivities; i++)
    	printf(" %s", activitynames[i]);
    printf("\n\n");

    printf("Prompt activity labels (%d):", NumPromptActivities);
    for (i=0; i<NumPromptActivities; i++)
    	printf(" %s", promptactivitynames[i]);
    printf("\n");

    printf("*****************************\n\n");
}


void PrintActivityInstances (int startActivity, int endActivity)
{
	for (int i = startActivity; i <= endActivity; i++)
	{
		printf ("Activity: %s (%d instances)\n", activitynames[i], numdataperactivity[i]);
		/*
		for (int j = 0; j < numdataperactivity[i]; j++)
		{
			PrintFeatures (adindices[i][j]);
		}
		printf ("\n");
		*/
	}
}


void PrintFeatures (int dataIndex)
{
	time_t startTime = data[dataIndex][FEATURE_START_TIME];
	time_t endTime = data[dataIndex][FEATURE_END_TIME];
	struct tm *tms;

	tms = localtime (&startTime);
	printf ("  Start %04d-%02d-%02d %02d:%02d:%02d, ",
			tms->tm_year + 1900, tms->tm_mon + 1, tms->tm_mday, tms->tm_hour, tms->tm_min, tms->tm_sec);
	tms = localtime (&endTime);
	printf ("End %04d-%02d-%02d %02d:%02d:%02d\n",
			tms->tm_year + 1900, tms->tm_mon + 1, tms->tm_mday, tms->tm_hour, tms->tm_min, tms->tm_sec);
}


void PrompterUsage ()
{
	printf ("Usage: prompter -learn <configuration_file>\n");
	printf ("       prompter -monitor <configuration_file>\n");
	printf ("       prompter -generate <configuration_file> <start_data> <start_time> <end_date> <end_time>\n");
	printf ("       prompter -test <configuration_file>\n");
}


/* Read and store the event data from the input files.
 * Activities are defined as a contiguous set of events with the same activity label.
 * For each such activity, its start and end time are stored with the activity. */
void ReadData()
{
	FILE *fp;
	char *cptr, buffer[MAX_LINE_LEN], eventDate[MAXSTR], eventTime[MAXSTR], lowlevelsensorid[MAXSTR];
	char highlevelsensorid[MAXSTR], sensorstatus[MAXSTR], activityLabel[MAXSTR];
	int numfile, activityNum, currentActivityNum;
	int features[NUM_FEATURES];

	NumData = 0;
	for (numfile=0; numfile<NumFiles; numfile++)  // Process each data file
	{
		printf("Reading data file %s...", filenames[numfile]);
		fp = fopen(filenames[numfile], "r");
		if (fp == NULL)
		{
			printf("Unable to open file\n");
			exit(1);
		}
		printf ("\n");
		activityNum = -1;
		currentActivityNum = -1;
		cptr = fgets(buffer, MAX_LINE_LEN, fp);
		while (cptr != NULL)  // Process each sensor event in the file
		{
			sscanf(buffer, "%s %s %s %s %s %s", eventDate, eventTime, highlevelsensorid, lowlevelsensorid, sensorstatus, activityLabel);
			activityNum = FindActivity (activityLabel);
			if (activityNum != currentActivityNum)
			{
				if (currentActivityNum >= 0) // a known activity
				{
					AddData (currentActivityNum, features);
				}
				// reset start/end time
				SetStartTime (eventDate, eventTime, features);
				SetEndTime (eventDate, eventTime, features);
				currentActivityNum = activityNum;
			} else {
				// advance end time
				SetEndTime (eventDate, eventTime, features);
			}
			cptr = fgets(buffer, MAX_LINE_LEN, fp);
			if (cptr == NULL) // end of file, process last activity
			{	
				if (currentActivityNum >= 0) // a known activity
				{
					AddData (currentActivityNum, features);
				}
			}
		}
		fclose(fp);
	}
	printf ("Read %d activity instances\n", NumData);
}


// Add prompter-specific events (year, month, week, day, hour, month of year, day of week, hour of day) to data.
// Note: these will only be used as possible relative activities (no patterns learned for these).

void AddPrompterEvents ()
{
	time_t currentTime, startTime, endTime;
	struct tm *localTime;
	int features[NUM_FEATURES];

	printf ("Adding prompter events...\n");
	startTime = DataStartTime;
	endTime = DataEndTime;
	AddBeforeAfterPrompterEvents (startTime, true);
	// Add prompter events during real sensor data time
	for (currentTime = startTime; currentTime <= endTime; currentTime++)
	{
		// Generate prompter activities for current time
		localTime = localtime (&currentTime);
		if ((localTime->tm_sec == 0) && (localTime->tm_min == 0))
		{
			SetFeatures (features, currentTime, currentTime);
			AddData (NumActivities + PROMPTER_ACTIVITY_HOUR, features);
			AddData (NumActivities + PROMPTER_ACTIVITY_HOUR_00 + localTime->tm_hour, features);
			if (localTime->tm_hour == 0)
			{
				AddData (NumActivities + PROMPTER_ACTIVITY_DAY, features);
				AddData (NumActivities + PROMPTER_ACTIVITY_SUNDAY + localTime->tm_wday, features);
				if (localTime->tm_wday == DAY_SUNDAY)
				{
					AddData (NumActivities + PROMPTER_ACTIVITY_WEEK, features);
				}
				if (localTime->tm_mday == 1)
				{
					AddData (NumActivities + PROMPTER_ACTIVITY_MONTH, features);
					AddData (NumActivities + PROMPTER_ACTIVITY_JANUARY + localTime->tm_mon, features);
					if (localTime->tm_mon == 0)
					{
						AddData (NumActivities + PROMPTER_ACTIVITY_YEAR, features);
					}
				}
			}
		}
	}
	AddBeforeAfterPrompterEvents (endTime, false);
}


// Add one of each prompter events just before (before = true) or after (before = false) real sensor data time.
// The currentTime is the start time (before = true) or end time (before = false) of the real sensor data.
// Note that we ensure that prompter events are added to the data in temporal order.
void AddBeforeAfterPrompterEvents (time_t currentTime, bool before)
{
	int i, increment, start, end;
	bool promptActivityUsed[NUM_PROMPTER_ACTIVITIES];
	int promptActivityNum[NUM_PROMPTER_ACTIVITIES];
	time_t promptActivityTime[NUM_PROMPTER_ACTIVITIES];
	int numPromptActivities;
	int features[NUM_FEATURES];
	struct tm *localTime;

	increment = 1;
	if (before) increment = -1;

	for (i = 0; i < NUM_PROMPTER_ACTIVITIES; i++)
	{
		promptActivityUsed[i] = false;
		promptActivityNum[i] = -1;
		promptActivityTime[i] = 0;
	}
	numPromptActivities = 0;
	while (numPromptActivities < NUM_PROMPTER_ACTIVITIES)
	{
		currentTime += increment;
		localTime = localtime (&currentTime);
		if ((localTime->tm_sec == 0) && (localTime->tm_min == 0))
		{
			if (! promptActivityUsed[PROMPTER_ACTIVITY_HOUR])
			{
				promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_HOUR;
				promptActivityTime[numPromptActivities] = currentTime;
				promptActivityUsed[PROMPTER_ACTIVITY_HOUR] = true;
				numPromptActivities++;
			}
			if (! promptActivityUsed[PROMPTER_ACTIVITY_HOUR_00 + localTime->tm_hour])
			{
				promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_HOUR_00 + localTime->tm_hour;
				promptActivityTime[numPromptActivities] = currentTime;
				promptActivityUsed[PROMPTER_ACTIVITY_HOUR_00 + localTime->tm_hour] = true;
				numPromptActivities++;
			}
			if (localTime->tm_hour == 0)
			{
				if (! promptActivityUsed[PROMPTER_ACTIVITY_DAY])
				{
					promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_DAY;
					promptActivityTime[numPromptActivities] = currentTime;
					promptActivityUsed[PROMPTER_ACTIVITY_DAY] = true;
					numPromptActivities++;
				}
				if (! promptActivityUsed[PROMPTER_ACTIVITY_SUNDAY + localTime->tm_wday])
				{
					promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_SUNDAY + localTime->tm_wday;
					promptActivityTime[numPromptActivities] = currentTime;
					promptActivityUsed[PROMPTER_ACTIVITY_SUNDAY + localTime->tm_wday] = true;
					numPromptActivities++;
				}
				if (localTime->tm_wday == DAY_SUNDAY)
				{
					if (! promptActivityUsed[PROMPTER_ACTIVITY_WEEK])
					{
						promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_WEEK;
						promptActivityTime[numPromptActivities] = currentTime;
						promptActivityUsed[PROMPTER_ACTIVITY_WEEK] = true;
						numPromptActivities++;
					}
				}
				if (localTime->tm_mday == 1)
				{
					if (! promptActivityUsed[PROMPTER_ACTIVITY_MONTH])
					{
						promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_MONTH;
						promptActivityTime[numPromptActivities] = currentTime;
						promptActivityUsed[PROMPTER_ACTIVITY_MONTH] = true;
						numPromptActivities++;
					}
					if (! promptActivityUsed[PROMPTER_ACTIVITY_JANUARY + localTime->tm_mon])
					{
						promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_JANUARY + localTime->tm_mon;
						promptActivityTime[numPromptActivities] = currentTime;
						promptActivityUsed[PROMPTER_ACTIVITY_JANUARY + localTime->tm_mon] = true;
						numPromptActivities++;
					}
					if (localTime->tm_mon == 0)
					{
						if (! promptActivityUsed[PROMPTER_ACTIVITY_YEAR])
						{
							promptActivityNum[numPromptActivities] = PROMPTER_ACTIVITY_YEAR;
							promptActivityTime[numPromptActivities] = currentTime;
							promptActivityUsed[PROMPTER_ACTIVITY_YEAR] = true;
							numPromptActivities++;
						}
					}
				}
			}
		}
	}
	// First ensure prompter events added in correct chronological order, then add them
	start = 0;
	end = NUM_PROMPTER_ACTIVITIES;
	if (before)
	{
		start = NUM_PROMPTER_ACTIVITIES - 1;
		end = -1;
	}
	for (i = start; i != end; i += increment)
	{
		SetFeatures (features, promptActivityTime[i], promptActivityTime[i]);
		AddData (NumActivities + promptActivityNum[i], features);
	}
}


void SetFeatures (int* features, int startTime, int endTime)
{
	features[FEATURE_START_TIME] = startTime;
	features[FEATURE_END_TIME] = endTime;
}


// Set the start time in features to the given year, month, day, hour, minute and second.
void SetStartTime (char *eventDate, char *eventTime, int *features)
{
	char tempStr[MAXSTR];
	struct tm timeStruct;

	// Compute year
	tempStr[0] = eventDate[0];
	tempStr[1] = eventDate[1];
	tempStr[2] = eventDate[2];
	tempStr[3] = eventDate[3];
	tempStr[4] = '\0';
	timeStruct.tm_year = atoi (tempStr) - 1900;
	// Compute month of year (1-12)
	tempStr[0] = eventDate[5];
	tempStr[1] = eventDate[6];
	tempStr[2] = '\0';
	timeStruct.tm_mon = atoi (tempStr) - 1;
	// Compute day of month (1-31)
	tempStr[0] = eventDate[8];
	tempStr[1] = eventDate[9];
	tempStr[2] = '\0';
	timeStruct.tm_mday = atoi (tempStr);
	// Compute hour (0-23)
	tempStr[0] = eventTime[0];
	tempStr[1] = eventTime[1];
	tempStr[2] = '\0';
	timeStruct.tm_hour = atoi (tempStr);
	// Compute minute of hour (0-59)
	tempStr[0] = eventTime[3];
	tempStr[1] = eventTime[4];
	tempStr[2] = '\0';
	timeStruct.tm_min = atoi (tempStr);
	// Compute second of minute (0-59), ignoring fractional seconds
	tempStr[0] = eventTime[6];
	tempStr[1] = eventTime[7];
	tempStr[2] = '\0';
	timeStruct.tm_sec = atoi (tempStr);

	timeStruct.tm_isdst = -1; // let mktime divine daylight savings
	features[FEATURE_START_TIME] = mktime (&timeStruct);
}


// Set the end time in features to the given year, month, day, hour, minute, and second.
void SetEndTime (char *eventDate, char *eventTime, int *features)
{
	char tempStr[MAXSTR];
	struct tm timeStruct;

	// Compute year
	tempStr[0] = eventDate[0];
	tempStr[1] = eventDate[1];
	tempStr[2] = eventDate[2];
	tempStr[3] = eventDate[3];
	tempStr[4] = '\0';
	timeStruct.tm_year = atoi (tempStr) - 1900;
	// Compute month of year (1-12)
	tempStr[0] = eventDate[5];
	tempStr[1] = eventDate[6];
	tempStr[2] = '\0';
	timeStruct.tm_mon = atoi (tempStr) - 1;
	// Compute day of month (1-31)
	tempStr[0] = eventDate[8];
	tempStr[1] = eventDate[9];
	tempStr[2] = '\0';
	timeStruct.tm_mday = atoi (tempStr);
	// Compute hour (0-23)
	tempStr[0] = eventTime[0];
	tempStr[1] = eventTime[1];
	tempStr[2] = '\0';
	timeStruct.tm_hour = atoi (tempStr);
	// Compute minute of hour (0-59)
	tempStr[0] = eventTime[3];
	tempStr[1] = eventTime[4];
	tempStr[2] = '\0';
	timeStruct.tm_min = atoi (tempStr);
	// Compute second of minute (0-59), ignoring fractional seconds
	tempStr[0] = eventTime[6];
	tempStr[1] = eventTime[7];
	tempStr[2] = '\0';
	timeStruct.tm_sec = atoi (tempStr);

	timeStruct.tm_isdst = -1; // let mktime divine daylight savings
	features[FEATURE_END_TIME] = mktime (&timeStruct);
}


// Returns start time from given feature vector in seconds since Epoch
time_t GetStartTime (int *features)
{
	return features[FEATURE_START_TIME];
}


// Returns end time from given feature vector in seconds since Epoch
time_t GetEndTime (int *features)
{
	return features[FEATURE_END_TIME];
}


// Add feature array to data array and to associated activity
void AddData (int activityNum, int* features)
{
	int i;

	if (NumData == 0)
	{
		data = (int **) malloc(sizeof(int *));
		DataStartTime = GetStartTime (features);
		DataEndTime = GetEndTime (features);
	} else {
		data = (int **) realloc(data, sizeof(int *) * (NumData+1));
		if (GetStartTime (features) < DataStartTime)
			DataStartTime = GetStartTime (features);
		if (GetEndTime (features) > DataEndTime)
			DataEndTime = GetEndTime (features);
	}
	data[NumData] = (int *) malloc(sizeof(int) * NUM_FEATURES);
	for (i=0; i<NUM_FEATURES; i++)
		data[NumData][i] = features[i];
	AddDataToActivity (activityNum, NumData);
	NumData++;
}


// Add the event data index to the activity list.
void AddDataToActivity (int activityNum, int eventIndex)
{
	if (numdataperactivity[activityNum] > 0)
	{
		adindices[activityNum] = (int *) realloc(adindices[activityNum], sizeof(int) * (numdataperactivity[activityNum]+1));
	} else {
		adindices[activityNum] = (int *) malloc(sizeof(int));
	}
	adindices[activityNum][numdataperactivity[activityNum]] = eventIndex;
	numdataperactivity[activityNum]++;
}


void ReadActivityPatterns ()
{
    FILE* fp;
    char *cptr, buffer[MAX_LINE_LEN], *tok;
    int i, activityNum, relativeActivityNum;
    int mean, stddev;

    printf ("Reading activity patterns from file %s...", patternfname);
    fp = fopen (patternfname, "r");
    if (fp == NULL)
    {
        printf("Unable to read pattern file\n");
        exit(1);
    }
    printf ("\n");
    cptr = fgets (buffer, MAX_LINE_LEN, fp);
    while (cptr != NULL)                // Process each line of the pattern file
    {
        tok = strtok(cptr, " \n");		// Process each string separated by " "
        activityNum = FindActivity (tok);
        if (activityNum < 0)
        {
        	printf ("  Unknown activity in pattern file: %s\n", tok);
        } else {
            tok = strtok (NULL, " \n");
            while (tok != NULL)
            {
            	if (strcmp (tok, "NO_PATTERN") != 0)
            	{
            		relativeActivityNum = FindActivity (tok);
            		tok = strtok (NULL, " \n");
            		mean = atoi (tok);
            		tok = strtok (NULL, " \n");
            		stddev = atoi (tok);
            		if (relativeActivityNum < 0)
            		{
            			printf ("  Unknown relative activity in pattern file: %s\n", tok);
            		} else {
            			// Add pattern to activity's pattern array
            			i = NumPatternsPerActivity[activityNum];
            			if (i == 0)
            				ActivityPatterns[activityNum] = (ActivityPattern*) malloc (sizeof(ActivityPattern));
            			else ActivityPatterns[activityNum] =
            					(ActivityPattern*) realloc (ActivityPatterns[activityNum], (sizeof(ActivityPattern) * (i + 1)));
            			ActivityPatterns[activityNum][i].relativeActivity = relativeActivityNum;
            			ActivityPatterns[activityNum][i].pValue = 0.0;
            			ActivityPatterns[activityNum][i].mean = mean;
            			ActivityPatterns[activityNum][i].stddev = stddev;
            			ActivityPatterns[activityNum][i].mode = PATTERN_MODE_INACTIVE;
            			ActivityPatterns[activityNum][i].time = 0;
            			NumPatternsPerActivity[activityNum]++;
            		}
            	}
            	tok = strtok (NULL, " \n");
            }
         }
        cptr = fgets (buffer, MAX_LINE_LEN, fp);  // Read next line
	}
	fclose (fp);
	if (OutputLevel > 0)
	{
		PrintActivityPatterns (stdout);
	}
}


// Write activity patterns to file specified in configuration
void WriteActivityPatterns ()
{
	FILE* fp = fopen (patternfname, "w");

	printf ("Writing activity patterns to file %s...", patternfname);
	if (fp == NULL)
	{
		printf ("Unable to open file.\n");
	} else {
		printf ("\n");
		PrintActivityPatterns (fp);
	}
	fclose (fp);
}


void PrintActivityPatterns (FILE* fp)
{
	int activity, pattern;

	for (activity = 0; activity < NumActivities; activity++)
	{
		fprintf (fp, "%s", activitynames[activity]);
		if (NumPatternsPerActivity[activity] > 0)
		{
			for (pattern = 0; pattern < NumPatternsPerActivity[activity]; pattern++)
			{
				fprintf (fp, " %s %d %d", activitynames[ActivityPatterns[activity][pattern].relativeActivity],
						ActivityPatterns[activity][pattern].mean, ActivityPatterns[activity][pattern].stddev);
			}
		} else {
			fprintf (fp, " NO_PATTERN");
		}
		fprintf (fp, "\n");
	}
}


// For each input activity A, try to find a pattern consisting of a relative activity RA and
// the mean and standard deviation of the duration between the occurrence of RA and the subsequent
// occurrence of A. See the user's manual for a more detailed explanation of how activity patterns
// are learned.
void LearnActivityPatterns ()
{
	int activity, relativeActivity;
	int bestRelativeActivity;
	int numA, numRA, numAafterRA;
	int mean, stdDev, bestRelativeActivityMean, bestRelativeActivityStdDev;
	float pValue, bestRelativeActivityPValue;
	bool semiPeriodicBetter;
	int i;

	printf ("Learning activity patterns...\n");
	for (activity = 0; activity < NumActivities; activity++)
	{
		bestRelativeActivity = -1;
		bestRelativeActivityMean = 0;
		bestRelativeActivityStdDev = 0;
		bestRelativeActivityPValue = 0.0;
		numA = numdataperactivity[activity];
		if (numA > 1)
		{
			// Consider input activities and prompter activities as possible relative activities
			for (relativeActivity = 0; relativeActivity < (NumActivities + NUM_PROMPTER_ACTIVITIES); relativeActivity++)
			{
				if (relativeActivity != activity)
				{
					pValue = 0.0;
					numRA = numdataperactivity[relativeActivity];
					if (numRA > 1)
					{
						numAafterRA = ComputeActivityStatistics (activity, relativeActivity, relativeActivity, &mean, &stdDev);
						if (numAafterRA > 0)
						{
							if (stdDev == 0)
							{
								pValue = numAafterRA * sqrt(numAafterRA) / numRA;
							} else {
								pValue = numAafterRA * sqrt(numAafterRA) / (numRA * ((float) stdDev));
							}
							if ((numRA - numAafterRA) > 0) pValue = pValue / sqrt (numRA - numAafterRA);
							if (mean != 0) pValue = pValue / sqrt (mean);
						}
					}
					if ((bestRelativeActivity == -1) ||
						(pValue > bestRelativeActivityPValue) ||
						((pValue == bestRelativeActivityPValue) && (mean < bestRelativeActivityMean)))
					{
						bestRelativeActivityPValue = pValue;
						bestRelativeActivity = relativeActivity;
						bestRelativeActivityMean = mean;
						bestRelativeActivityStdDev = stdDev;
					}
				}
			}

			if (OutputLevel > 1) // DEBUG
			{
				printf ("  Best single pattern for activity %s:", activitynames[activity]);
				printf (" %s + %d +/- %d (P=%f)\n", activitynames[bestRelativeActivity],
						bestRelativeActivityMean, bestRelativeActivityStdDev, bestRelativeActivityPValue);
			}

			semiPeriodicBetter = ConsiderSemiPeriodicPatterns (activity, bestRelativeActivityPValue);
		}
		// Store best single pattern
		if ((! semiPeriodicBetter) && (bestRelativeActivity >= 0))
		{
			ActivityPatterns[activity] = (ActivityPattern*) malloc (sizeof(ActivityPattern));
			ActivityPatterns[activity][0].relativeActivity = bestRelativeActivity;
			ActivityPatterns[activity][0].pValue = bestRelativeActivityPValue;
			ActivityPatterns[activity][0].mean = bestRelativeActivityMean;
			ActivityPatterns[activity][0].stddev = bestRelativeActivityStdDev;
			ActivityPatterns[activity][0].mode = PATTERN_MODE_INACTIVE;
   			ActivityPatterns[activity][0].time = 0;
			NumPatternsPerActivity[activity] = 1;
		}
		if (OutputLevel > 0)
		{
			if (NumPatternsPerActivity[activity] > 0)
			{
				printf ("Pattern for %s:", activitynames[activity]);
				for (i = 0; i < NumPatternsPerActivity[activity]; i++)
				{
					printf (" %s + %d +/- %d (P=%f)",
							activitynames[ActivityPatterns[activity][i].relativeActivity],
							ActivityPatterns[activity][i].mean, ActivityPatterns[activity][i].stddev,
							ActivityPatterns[activity][i].pValue);
				}
				printf ("\n");
			} else {
				printf ("Pattern for %s: no pattern (occurrences = %d)\n", activitynames[activity], numA);
			}
		}
	}
}


// Considers semi-periodic patterns (some of the months of a year, some of the days of a week,
// some of the hours of a day) for given activity. Value of periodic pattern must be better
// than given value of best single-activity pattern. If semi-periodic pattern better, then
// store in ActivityPatterns array and return true, otherwise return false.
bool ConsiderSemiPeriodicPatterns (int activity, float bestPValue)
{
	bool semiPeriodicBetter = false;
	int numRA, numAafterRA;
	int relativeActivity, nextRelativeActivity;
	float pValue;
	float semiMonthlyPValue, semiDailyPValue, semiHourlyPValue;
	float *pValues;
	int mean, stddev;
	int *means, *stddevs;
	int i;
	int numPatterns;
	int bestRelativeActivityStart, bestRelativeActivityEnd;
	int *numAafterRAs;
	int maxNumAafterRA;
	int bestNumAafterRA;

	pValues = (float *) malloc (sizeof(float) * (NumActivities + NUM_PROMPTER_ACTIVITIES));
	means = (int *) malloc (sizeof(int) * (NumActivities + NUM_PROMPTER_ACTIVITIES));
	stddevs = (int *) malloc (sizeof(int) * (NumActivities + NUM_PROMPTER_ACTIVITIES));
	numAafterRAs = (int *) malloc (sizeof(int) * (NumActivities + NUM_PROMPTER_ACTIVITIES));
	for (i = 0; i < (NumActivities + NUM_PROMPTER_ACTIVITIES); i++)
	{
		pValues[i] = 0.0;
		means[i] = 0;
		stddevs[i] = 0;
		numAafterRAs[i] = 0;
	}

	// Consider some months of the year
	// -- Compute pValues for each month individually
	maxNumAafterRA = 0;
	for (relativeActivity = (NumActivities + PROMPTER_ACTIVITY_JANUARY);
		 relativeActivity <= (NumActivities + PROMPTER_ACTIVITY_DECEMBER);
		 relativeActivity++)
	{
		pValue = 0.0;
		numRA = numdataperactivity[relativeActivity];
		if (numRA > 1)
		{
			nextRelativeActivity = relativeActivity + 1;
			if (nextRelativeActivity > (NumActivities + PROMPTER_ACTIVITY_DECEMBER))
			{
				nextRelativeActivity = (NumActivities + PROMPTER_ACTIVITY_JANUARY);
			}
			numAafterRA = ComputeActivityStatistics (activity, relativeActivity, nextRelativeActivity, &mean, &stddev);
			means[relativeActivity] = mean;
			stddevs[relativeActivity] = stddev;
			if (numAafterRA > 0)
			{
				if (stddev == 0)
				{
					pValue = numAafterRA * sqrt(numAafterRA) / numRA;
				} else {
					pValue = numAafterRA * sqrt(numAafterRA) / (numRA * ((float) stddev));
				}
				if ((numRA - numAafterRA) > 0) pValue = pValue / sqrt (numRA - numAafterRA);
				if (mean != 0) pValue = pValue / sqrt (mean);
				numAafterRAs[relativeActivity] = numAafterRA;
				if (numAafterRA > maxNumAafterRA) maxNumAafterRA = numAafterRA;
			}
		}
		pValues[relativeActivity] = pValue;

		if (OutputLevel > 1) // DEBUG
		{
			printf ("  Month-of-year pattern for %s: ", activitynames[activity]);
			printf ("%s + %d +/- %d (P=%f, numRA=%d, numAafterRA=%d)\n", activitynames[relativeActivity], mean, stddev, pValue,
					numRA, numAafterRA);
		}
	}
	// -- Add up pValues for all relative activities whose frequency is high
	semiMonthlyPValue = 0.0;
	for (relativeActivity = (NumActivities + PROMPTER_ACTIVITY_JANUARY);
		 relativeActivity <= (NumActivities + PROMPTER_ACTIVITY_DECEMBER);
		 relativeActivity++)
	{
		if (numAafterRAs[relativeActivity] > (maxNumAafterRA / 2))
		{
			semiMonthlyPValue += pValues[relativeActivity];
		}
	}
	if (semiMonthlyPValue > bestPValue)
	{
		bestPValue = semiMonthlyPValue;
		bestNumAafterRA = maxNumAafterRA / 2;
		bestRelativeActivityStart = (NumActivities + PROMPTER_ACTIVITY_JANUARY);
		bestRelativeActivityEnd = (NumActivities + PROMPTER_ACTIVITY_DECEMBER);
		semiPeriodicBetter = true;
	}

	// Consider some days of the week
	// -- Compute pValues for each day individually
	maxNumAafterRA = 0;
	for (relativeActivity = (NumActivities + PROMPTER_ACTIVITY_SUNDAY);
		 relativeActivity <= (NumActivities + PROMPTER_ACTIVITY_SATURDAY);
		 relativeActivity++)
	{
		pValue = 0.0;
		numRA = numdataperactivity[relativeActivity];
		if (numRA > 1)
		{
			nextRelativeActivity = relativeActivity + 1;
			if (nextRelativeActivity > (NumActivities + PROMPTER_ACTIVITY_SATURDAY))
			{
				nextRelativeActivity = (NumActivities + PROMPTER_ACTIVITY_SUNDAY);
			}
			numAafterRA = ComputeActivityStatistics (activity, relativeActivity, nextRelativeActivity, &mean, &stddev);
			means[relativeActivity] = mean;
			stddevs[relativeActivity] = stddev;
			if (numAafterRA > 0)
			{
				if (stddev == 0)
				{
					pValue = numAafterRA * sqrt(numAafterRA) / (numRA);
				} else {
					pValue = numAafterRA * sqrt(numAafterRA) / (numRA * ((float) stddev));
				}
				if ((numRA - numAafterRA) > 0) pValue = pValue / sqrt (numRA - numAafterRA);
				if (mean != 0) pValue = pValue / sqrt (mean);
				numAafterRAs[relativeActivity] = numAafterRA;
				if (numAafterRA > maxNumAafterRA) maxNumAafterRA = numAafterRA;
			}
		}
		pValues[relativeActivity] = pValue;

		if (OutputLevel > 1) // DEBUG
		{
			printf ("  Day-of-week pattern for %s: ", activitynames[activity]);
			printf ("%s + %d +/- %d (P=%f, numRA=%d, numAafterRA=%d)\n", activitynames[relativeActivity], mean, stddev, pValue,
					numRA, numAafterRA);
		}
	}
	//-- Add up pValues for all relative activities whose frequency is high
	semiDailyPValue = 0.0;
	for (relativeActivity = (NumActivities + PROMPTER_ACTIVITY_SUNDAY);
		 relativeActivity <= (NumActivities + PROMPTER_ACTIVITY_SATURDAY);
		 relativeActivity++)
	{
		if (numAafterRAs[relativeActivity] > (maxNumAafterRA / 2))
		{
			semiDailyPValue += pValues[relativeActivity];
		}
	}
	if (semiDailyPValue > bestPValue)
	{
		bestPValue = semiDailyPValue;
		bestNumAafterRA = maxNumAafterRA / 2;
		bestRelativeActivityStart = (NumActivities + PROMPTER_ACTIVITY_SUNDAY);
		bestRelativeActivityEnd = (NumActivities + PROMPTER_ACTIVITY_SATURDAY);
		semiPeriodicBetter = true;
	}

	// Consider some hours of the day
	// -- Compute pValues for each hour individually
	maxNumAafterRA = 0;
	for (relativeActivity = (NumActivities + PROMPTER_ACTIVITY_HOUR_00);
		 relativeActivity <= (NumActivities + PROMPTER_ACTIVITY_HOUR_23);
		 relativeActivity++)
	{
		pValue = 0.0;
		numRA = numdataperactivity[relativeActivity];
		if (numRA > 1)
		{
			nextRelativeActivity = relativeActivity + 1;
			if (nextRelativeActivity > (NumActivities + PROMPTER_ACTIVITY_HOUR_23))
			{
				nextRelativeActivity = (NumActivities + PROMPTER_ACTIVITY_HOUR_00);
			}
			numAafterRA = ComputeActivityStatistics (activity, relativeActivity, nextRelativeActivity, &mean, &stddev);
			means[relativeActivity] = mean;
			stddevs[relativeActivity] = stddev;
			if (numAafterRA > 0)
			{
				if (stddev == 0)
				{
					pValue = numAafterRA * sqrt(numAafterRA) / (numRA);
				} else {
					pValue = numAafterRA * sqrt(numAafterRA) / (numRA * ((float) stddev));
				}
				if ((numRA - numAafterRA) > 0) pValue = pValue / sqrt (numRA - numAafterRA);
				if (mean != 0) pValue = pValue / sqrt (mean);
				numAafterRAs[relativeActivity] = numAafterRA;
				if (numAafterRA > maxNumAafterRA) maxNumAafterRA = numAafterRA;
			}
		}
		pValues[relativeActivity] = pValue;

		if (OutputLevel > 1) // DEBUG
		{
			printf ("  Hour-of-day pattern for %s: ", activitynames[activity]);
			printf ("%s + %d +/- %d (P=%f, numRA=%d, numAafterRA=%d)\n", activitynames[relativeActivity], mean, stddev, pValue,
					numRA, numAafterRA);
		}
	}
	//-- Add up pValues for all relative activities whose pValues are greater than the mean pValue
	semiHourlyPValue = 0.0;
	for (relativeActivity = (NumActivities + PROMPTER_ACTIVITY_HOUR_00);
		 relativeActivity <= (NumActivities + PROMPTER_ACTIVITY_HOUR_23);
		 relativeActivity++)
	{
		if (numAafterRAs[relativeActivity] > (maxNumAafterRA / 2))
		{
			semiHourlyPValue += pValues[relativeActivity];
		}
	}
	if (semiHourlyPValue > bestPValue)
	{
		bestPValue = semiHourlyPValue;
		bestNumAafterRA = maxNumAafterRA / 2;
		bestRelativeActivityStart = (NumActivities + PROMPTER_ACTIVITY_HOUR_00);
		bestRelativeActivityEnd = (NumActivities + PROMPTER_ACTIVITY_HOUR_23);
		semiPeriodicBetter = true;
	}

	// Define appropriate activity patterns
	if (semiPeriodicBetter)
	{
		numPatterns = NumPatternsPerActivity[activity];
		for (i = bestRelativeActivityStart; i <= bestRelativeActivityEnd; i++)
		{
			if (numAafterRAs[i] > bestNumAafterRA)
			{
				if (numPatterns == 0)
				{
					ActivityPatterns[activity] = (ActivityPattern*) malloc (sizeof(ActivityPattern));
				} else {
					ActivityPatterns[activity] = (ActivityPattern*) realloc (ActivityPatterns[activity],
							(sizeof(ActivityPattern) * (numPatterns + 1)));
				}
				ActivityPatterns[activity][numPatterns].relativeActivity = i;
				ActivityPatterns[activity][numPatterns].pValue = pValues[i];
				ActivityPatterns[activity][numPatterns].mean = means[i];
				ActivityPatterns[activity][numPatterns].stddev = stddevs[i];
				ActivityPatterns[activity][numPatterns].mode = PATTERN_MODE_INACTIVE;
				ActivityPatterns[activity][numPatterns].time = 0;
				numPatterns++;
			}
		}
		NumPatternsPerActivity[activity] = numPatterns;
	}

	free (pValues);
	free (means);
	free (stddevs);

	return semiPeriodicBetter;
}


// Compute the mean and standard deviation in seconds of the occurrences of activity between occurrences of
// relativeActivity and nextRelativeActivity. Return the number of times activity occurs at least once between
// relativeActivity and nextRelativeActivity.
int ComputeActivityStatistics (int activity, int relativeActivity, int nextRelativeActivity, int *mean, int *stddev)
{
	float localMean = 0.0;
	float localStdDev = 0.0;
	float* localData = NULL;
	int count = 0;
	int freq = 0;
	bool foundOne;
	time_t aOcc, raOcc1, raOcc2;
	int actIndex, relActIndex, nextRelActIndex;

	localData = (float *) malloc (sizeof(float) * numdataperactivity[activity]);
	for (actIndex = 0; actIndex < numdataperactivity[activity]; actIndex++)
	{
		localData[actIndex] = -1;
	}
	actIndex = 0;
	nextRelActIndex = 0;
	raOcc2 = 0;
	aOcc = 0;
	for (relActIndex = 0; relActIndex < numdataperactivity[relativeActivity]; relActIndex++)
	{
		foundOne = false;
		raOcc1 = GetStartTime (data[adindices[relativeActivity][relActIndex]]);
		// Find next occurrence of nextRelativeActivity after this relativeActivity
		while ((raOcc2 <= raOcc1) && (nextRelActIndex < numdataperactivity[nextRelativeActivity]))
		{
			raOcc2 = GetStartTime (data[adindices[nextRelativeActivity][nextRelActIndex]]);
			nextRelActIndex++;
		}
		if (raOcc2 > raOcc1) // found a nextRelativeActivity
		{
			// Find next occurrence of activity after this relativeActivity
			while ((aOcc <= raOcc1) && (actIndex < numdataperactivity[activity]))
			{
				aOcc = GetStartTime (data[adindices[activity][actIndex]]);
				actIndex++;
			}
			if (aOcc > raOcc1) // found an activity
			{
				while ((aOcc < raOcc2) && ((actIndex - 1) < numdataperactivity[activity]))
				{
					localData[actIndex-1] = aOcc - raOcc1;
					foundOne = true;
					if (actIndex < numdataperactivity[activity])
					{
						aOcc = GetStartTime (data[adindices[activity][actIndex]]);
					}
					actIndex++;
				}
			}
		}
		if (foundOne)
		{
			freq++;
		}
	}
	// Compute mean and standard deviation
	count = 0;
	for (actIndex = 0; actIndex < numdataperactivity[activity]; actIndex++)
	{
		if (localData[actIndex] >= 0)
		{
			localMean += localData[actIndex];
			count++;
		}
	}
	if (count > 0)
	{
		localMean = localMean / count;
		for (actIndex = 0; actIndex < numdataperactivity[activity]; actIndex++)
		{
			if (localData[actIndex] >= 0)
			{
				localStdDev += ((localData[actIndex] - localMean) * (localData[actIndex] - localMean));
			}
		}
		localStdDev = sqrt (localStdDev / count);
	}
	free (localData);
	*mean = lroundf (localMean);
	*stddev = lroundf (localStdDev);
	return freq;
}


/* Find the index of a specified activity from the activity list */
int FindActivity (char *name)
{
    int i;

    for (i = 0; i < (NumActivities + NUM_PROMPTER_ACTIVITIES); i++)
    {
        if (strcmp(name, activitynames[i]) == 0)
            return i;
    }

    return(-1);
}


/* Find the index of a specified sensor from the sensor list */
int FindSensor (char *name)
{
    int i;

    for (i=0; i<NumSensors; i++)
        if (strcmp(name, sensornames[i]) == 0)
            return(i);

    return(-1);
}


// Allocate memory for prompter (called after ReadConfig)
void AllocateMemory ()
{
	int i;

    adindices = (int **) malloc (sizeof(int *) * (NumActivities + NUM_PROMPTER_ACTIVITIES));
    numdataperactivity = (int *) malloc (sizeof(int) * (NumActivities + NUM_PROMPTER_ACTIVITIES));
    for (i = 0; i < (NumActivities + NUM_PROMPTER_ACTIVITIES); i++)
    {
    	adindices[i] = NULL;
    	numdataperactivity[i] = 0;
    }
    ActivityPatterns = (ActivityPattern**) malloc (sizeof(ActivityPattern*) * NumActivities);
    NumPatternsPerActivity = (int *) malloc (sizeof(int) * NumActivities);
    for (i = 0; i < NumActivities; i++)
    {
    	ActivityPatterns[i] = NULL;
    	NumPatternsPerActivity[i] = 0;
    }
    ActivityStatistics = (ActivityStats *) malloc (sizeof(ActivityStats) * NumActivities);
    for (i = 0; i < NumActivities; i++)
    {
    	ActivityStatistics[i].count = 0;
    	ActivityStatistics[i].error = 0;
    	ActivityStatistics[i].truePositives = 0;
    	ActivityStatistics[i].falsePositives = 0;
    	ActivityStatistics[i].falseNegatives = 0;
    }
}


/* Free allocated memory. */
void FreeMemory()
{
    int i;

    free(patternfname);

    for (i=0; i<NumFiles; i++)
    	free(filenames[i]);
    free(filenames);

    for (i=0; i<NumSensors; i++)
    	free(sensornames[i]);
    free(sensornames);

    for (i=0; i<NumData; i++)
    	free(data[i]);
    free(data);

    for (i = 0; i < (NumActivities + NUM_PROMPTER_ACTIVITIES); i++)
    {
    	if (numdataperactivity[i] > 0)
    		free(adindices[i]);
    	free(activitynames[i]);
    }
    free(adindices);
    free(activitynames);
    free(numdataperactivity);
    for (i = 0; i < NumActivities; i++)
    {
    	free (ActivityPatterns[i]);
    }
    free (ActivityPatterns);
    free (NumPatternsPerActivity);
    free (ActivityStatistics);
}


// ---------------------------------
// prompter -monitor mode
// ---------------------------------

/*
 * Outline:
 *   Read promptable activity pattern file
 *     Pattern = activity, relative activity, mean delay, standard deviation
 *   Watch for events and their activity label
 *     Maintain timestamp of last occurrence of each activity
 *     If activity is a relative activity for a completed promptable activity
 *     Then mark promptable activity as pending
 *   Monitor for omission of promptable activities
 *     If pending promptable activity has not occurred within mean+1sd of relative activity
 *     Then issue third and final prompt; mark promptable activity as completed
 *     If pending promptable activity has not occurred within mean of relative activity
 *     Then issue second prompt (okay, ignore)
 *          If ignore, then mark promptable activity as completed
 *     If pending promptable activity has not occurred within mean-1sd of relative activity
 *     Then issue first prompt (okay, ignore)
 *          If ignore, then mark promptable activity as completed
 */


// Set up a connection with the CASAS middleware
void InitMonitorInterface()
{
   string userName = prompterUserName + "@" + serverName + "/onlyone";

   // Create a new CASAS object for Prompter
   cbot = new Casas("Prompter", userName, prompterPassword);

   // Subscribe to the activity recognition channel
   cbot->addSubscribeChannel (activityRecognitionChannel);
   cbot->setChanMsgCallback(activityRecognitionChannel, HandleAR);

   // Subscribe to the chime channel to get an event every minute
   cbot->addSubscribeChannel (chimeChannel);
   cbot->setChanMsgCallback (chimeChannel, HandleChime);

   // Publish to the prompt channel so that Prompter can post prompting events
   cbot->addPublishChannel(promptChannel);

   // Subscribe to the prompt channel so that the prompting events can be printed
   // This is for testing; usually a separate agent would handle prompt requests
   cbot->addSubscribeChannel(promptChannel);
   cbot->setChanMsgCallback(promptChannel, HandlePrompt);
}


// Process a sensor event that has been labeled with the activity
void HandleAR (map<string, string> packet)
{
	char stamp[MAXSTR], date[MAXSTR], time[MAXSTR], activityName[MAXSTR];
	int features[NUM_FEATURES];
	time_t activityTime;
	int activity;

	// Get time of activity
	strcpy(stamp, packet["stamp"].c_str());
	sscanf(stamp, "%s %s", date, time);
	SetStartTime (date, time, features);
	activityTime = GetStartTime (features);

	// Get activity
	strcpy (activityName, packet["label"].c_str());
	activity = FindActivity (activityName);
	if (activity >= 0)
	{
		UpdatePatternsByActivity (activity, activityTime);
	}
}


// Prompts sent from prompter are ignored. Responses to prompts are processed.
// A "cancel" response deactivates the activity's active prompts. An "okay"
// response is ignored.
void HandlePrompt (map<string, string> packet)
{
	int j, activity;
	string activityName;
	char actName[MAXSTR];

	if (packet["value"] == "cancel")
	{
		// Got a "cancel" response from the user; deactivate activity's active prompts
		cout << "Received Cancel response to prompt for activity: " << packet["label"] << endl;
		activityName = packet["label"];
		strcpy (actName, activityName.c_str());
		activity = FindActivity (actName);
		for (j = 0; j < NumPatternsPerActivity[activity]; j++)
		{
			ActivityPatterns[activity][j].mode = PATTERN_MODE_INACTIVE;
		}
	}
	if (packet["value"] == "okay")
	{
		cout << "Received Okay response to prompt for activity: " << packet["label"] << endl;
	}
}


void HandleChime (map<string, string> packet)
{
	time_t currentTime = time (NULL);
	struct tm *localTime = localtime (&currentTime);

	if (OutputLevel > 0)
	{
		cout << "Chime received: " << packet["message"] << endl;
		PrintPacket (packet);
	}

	if (packet["message"] == "1-min")
	{
		// Generate prompt activities as appropriate
		if (localTime->tm_min == 0)
		{
			UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_HOUR, currentTime);
			UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_HOUR_00 + localTime->tm_hour, currentTime);
			if (localTime->tm_hour == 0)
			{
				UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_DAY, currentTime);
				UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_SUNDAY + localTime->tm_wday, currentTime);
				if (localTime->tm_wday == DAY_SUNDAY)
				{
					UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_WEEK, currentTime);
				}
				if (localTime->tm_mday == 1)
				{
					UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_MONTH, currentTime);
					UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_JANUARY + localTime->tm_mon, currentTime);
					if (localTime->tm_mon == 0)
					{
						UpdatePatternsByActivity (NumActivities + PROMPTER_ACTIVITY_YEAR, currentTime);
					}
				}
			}
		}
		CheckPrompts();
	}
}


void PrintPacket (map<string, string> packet)
{
   map<string, string>::iterator it;
   for (it = packet.begin(); it != packet.end(); it++)
   {
      cout << "  " << it->first << " : " << it->second << endl;
   }
}


// Check if any activity needs to be prompted. Should be called at least once a minute
// and after calls to UpdatePatternsByActivity.
void CheckPrompts ()
{
	int i, j;
	int elapsedTime, meanTime, sdTime;
	time_t currentTime = time (NULL);

	for (i = 0; i < NumActivities; i++)
	{
		for (j = 0; j < NumPatternsPerActivity[i]; j++)
		{
			if (ActivityPatterns[i][j].mode != PATTERN_MODE_INACTIVE)
			{
				meanTime = ActivityPatterns[i][j].mean;
				sdTime = ActivityPatterns[i][j].stddev;
				elapsedTime = currentTime - ActivityPatterns[i][j].time;
				if ((elapsedTime > (meanTime + sdTime)) && (ActivityPatterns[i][j].mode == PATTERN_MODE_TWO_PROMPTS))
				{
					IssuePrompt (i, 3);
					ActivityPatterns[i][j].mode = PATTERN_MODE_INACTIVE;
				} else if ((elapsedTime > meanTime) && (ActivityPatterns[i][j].mode == PATTERN_MODE_ONE_PROMPT)) {
					IssuePrompt (i, 2);
					ActivityPatterns[i][j].mode = PATTERN_MODE_TWO_PROMPTS;
				} else if ((elapsedTime > (meanTime - sdTime)) && (ActivityPatterns[i][j].mode == PATTERN_MODE_ACTIVE)) {
					IssuePrompt (i, 1);
					ActivityPatterns[i][j].mode = PATTERN_MODE_ONE_PROMPT;
				}
			}
		}
	}
}


void IssuePrompt (int activity, int promptNumber)
{
	if (OutputLevel > 0)
	{
		cout << "Issuing prompt " << promptNumber << " for activity " << activitynames[activity] << endl;
	}
	stringstream ss;
	ss << promptNumber;
	cbot->issuePrompt ("Prompter", activitynames[activity], ss.str(), promptChannel);
}


// Check each pattern to see if given activity is the pattern's activity or relative activity,
// and if so, update accordingly. If we see a new occurrence of a pattern's relative activity
// while that pattern is active, then we do not restart the pattern.
void UpdatePatternsByActivity (int activity, time_t activityTime)
{
	int i, j;

	if (OutputLevel > 1) // DEBUG
	{
		cout << "Updating patterns: activity = " << activitynames[activity] << ", time = " << ctime (&activityTime);
	}

	for (i = 0; i < NumActivities; i++)
	{
		for (j = 0; j < NumPatternsPerActivity[i]; j++)
		{
			if (activity == i)
			{
				ActivityPatterns[i][j].mode = PATTERN_MODE_INACTIVE;
			}
			if (activity == ActivityPatterns[i][j].relativeActivity)
			{
				if (ActivityPatterns[i][j].mode == PATTERN_MODE_INACTIVE)
				{
					ActivityPatterns[i][j].mode = PATTERN_MODE_ACTIVE;
					ActivityPatterns[i][j].time = activityTime;
				}
			}
		}
	}
}


// Write synthetic sensor data to configuration data file consistent with activity patterns
// and occurring between the given start and end times.
void GenerateData (time_t startTime, time_t endTime)
{
	time_t currentTime;
	struct tm *localTime;
	FILE* fp;
	int i, j;

	printf ("Generating data to file %s...", filenames[0]);
	fp = fopen (filenames[0], "w");
	if (fp == NULL)
	{
		printf ("Unable to open data file: %s\n", filenames[0]);
		exit (1);
	}
	printf ("\n");
	for (currentTime = startTime; currentTime <= endTime; currentTime++)
	{
		// Generate prompter activities for current time
		localTime = localtime (&currentTime);
		if ((localTime->tm_sec == 0) && (localTime->tm_min == 0))
		{
			GenerateActivity (NumActivities + PROMPTER_ACTIVITY_HOUR, currentTime);
			GenerateActivity (NumActivities + PROMPTER_ACTIVITY_HOUR_00 + localTime->tm_hour, currentTime);
			if (localTime->tm_hour == 0)
			{
				GenerateActivity (NumActivities + PROMPTER_ACTIVITY_DAY, currentTime);
				GenerateActivity (NumActivities + PROMPTER_ACTIVITY_SUNDAY + localTime->tm_wday, currentTime);
				if (localTime->tm_wday == DAY_SUNDAY)
				{
					GenerateActivity (NumActivities + PROMPTER_ACTIVITY_WEEK, currentTime);
				}
				if (localTime->tm_mday == 1)
				{
					GenerateActivity (NumActivities + PROMPTER_ACTIVITY_MONTH, currentTime);
					GenerateActivity (NumActivities + PROMPTER_ACTIVITY_JANUARY + localTime->tm_mon, currentTime);
					if (localTime->tm_mon == 0)
					{
						GenerateActivity (NumActivities + PROMPTER_ACTIVITY_YEAR, currentTime);
					}
				}
			}
		}

		// Check if scheduled activities need to be executed
		for (i = 0; i < NumActivities; i++)
		{
			for (j = 0; j < NumPatternsPerActivity[i]; j++)
			{
				if ((ActivityPatterns[i][j].mode == PATTERN_MODE_ACTIVE) &&
					(ActivityPatterns[i][j].time == currentTime))
				{
					WriteActivity (fp, i, currentTime);
					GenerateActivity (i, currentTime);
					ActivityPatterns[i][j].mode = PATTERN_MODE_INACTIVE;
				}
			}
		}
	}
	fclose (fp);
}


// If given activity is a relative activity for an inactive pattern,
// then schedule pattern's activity after a delay determined by the
// pattern's mean and standard deviation.
void GenerateActivity (int activity, time_t currentTime)
{
	int i, j;
	int delayTime;

	for (i = 0; i < NumActivities; i++)
	{
		for (j = 0; j < NumPatternsPerActivity[i]; j++)
		{
			if ((ActivityPatterns[i][j].mode == PATTERN_MODE_INACTIVE) &&
				(ActivityPatterns[i][j].relativeActivity == activity))
			{
				delayTime = SampleNormal (ActivityPatterns[i][j].mean, ActivityPatterns[i][j].stddev);
				if (delayTime < 1) delayTime = 1;
				ActivityPatterns[i][j].time = currentTime + delayTime;
				ActivityPatterns[i][j].mode = PATTERN_MODE_ACTIVE;
			}
		}
	}
}


void WriteActivity (FILE *fp, int activity, time_t currentTime)
{
	struct tm *localTime;

	localTime = localtime (&currentTime);
	fprintf (fp, "%04d-%02d-%02d %02d:%02d:%02d.0 SENSOR SENSOR VALUE %s\n",
			 (localTime->tm_year + 1900), (localTime->tm_mon + 1), localTime->tm_mday,
			 localTime->tm_hour, localTime->tm_min, localTime->tm_sec,
			 activitynames[activity]);
}


// Return sample from normal distribution using Box-Muller method
int SampleNormal (int mean, int stddev)
{
	double u = ((double) random()) / ((double) RAND_MAX);
	double v = ((double) random()) / ((double) RAND_MAX);
	double z = sqrt (-2.0 * log(u)) * cos (2.0 * M_PI * v);
	double x = (z * ((double) stddev)) + ((double) mean);
	return lround (x);
}


// Compute and output statistics on the accuracy of the activity patterns when
// applied to the current data.
void TestData ()
{
	int i;
	time_t currentTime;
	int activityIndices[NumActivities + NUM_PROMPTER_ACTIVITIES];
	int totalCount, totalError, totalTruePositives, totalFalsePositives, totalFalseNegatives;
	int meanTPError;
	float accuracy;

	printf ("Testing activity patterns on data...\n");
	for (i = 0; i < (NumActivities + NUM_PROMPTER_ACTIVITIES); i++)
	{
		activityIndices[i] = 0;
	}
	for (currentTime = DataStartTime; currentTime <= DataEndTime; currentTime++)
	{
		// Process all sensor data and prompt events occurring at currentTime
		for (i = 0; i < (NumActivities + NUM_PROMPTER_ACTIVITIES); i++)
		{
			if (activityIndices[i] < numdataperactivity[i])
			{
				if (GetStartTime (data[adindices[i][activityIndices[i]]]) == currentTime)
				{
					EvaluatePatternsByActivity (i, currentTime);
					activityIndices[i]++;
				}
			}
		}
	}
	// Output statistics
	totalCount = 0;
	totalError = 0;
	totalTruePositives = 0;
	totalFalsePositives = 0;
	totalFalseNegatives = 0;
	for (i = 0; i < NumActivities; i++)
	{
		accuracy = 0.0;
		if (ActivityStatistics[i].count > 0)
			accuracy = ActivityStatistics[i].truePositives / ((float) ActivityStatistics[i].count);
		meanTPError = 0;
		if (ActivityStatistics[i].truePositives > 0)
			meanTPError = ActivityStatistics[i].error / ActivityStatistics[i].truePositives;
		printf ("Activity: %s\n", activitynames[i]);
		printf ("  Count = %d\n", ActivityStatistics[i].count);
		printf ("  Accuracy = %f\n", accuracy);
		printf ("  True positives = %d (mean error = %d seconds)\n", ActivityStatistics[i].truePositives, meanTPError);
		printf ("  False positives = %d\n", ActivityStatistics[i].falsePositives);
		printf ("  False negatives = %d\n", ActivityStatistics[i].falseNegatives);
		totalCount += ActivityStatistics[i].count;
		totalError += ActivityStatistics[i].error;
		totalTruePositives += ActivityStatistics[i].truePositives;
		totalFalsePositives += ActivityStatistics[i].falsePositives;
		totalFalseNegatives += ActivityStatistics[i].falseNegatives;
	}
	accuracy = 0.0;
	if (totalCount > 0)
		accuracy = totalTruePositives / ((float) totalCount);
	meanTPError = 0;
	if (totalTruePositives > 0)
		meanTPError = totalError / totalTruePositives;
	printf ("Activity: ALL\n");
	printf ("  Count = %d\n", totalCount);
	printf ("  Accuracy = %f\n", accuracy);
	printf ("  True positives = %d (mean error = %d seconds)\n", totalTruePositives, meanTPError);
	printf ("  False positives = %d\n", totalFalsePositives);
	printf ("  False negatives = %d\n", totalFalseNegatives);
}

// Check each pattern to see if given activity is the pattern's activity or relative activity,
// and if so, update accordingly.
void EvaluatePatternsByActivity (int activity, time_t activityTime)
{
	int i, j;
	bool predicted;
	time_t errorTime;

	if (OutputLevel > 1) // DEBUG
	{
		cout << "Evaluating patterns: activity = " << activitynames[activity] << ", time = " << ctime (&activityTime);
	}

	if (activity < NumActivities)
		ActivityStatistics[activity].count++;
	predicted = false;
	for (i = 0; i < NumActivities; i++)
	{
		for (j = 0; j < NumPatternsPerActivity[i]; j++)
		{
			if (activity == i)
			{
				// if this pattern is active (and no other pattern has predicted this activity), then true positive
				if ((! predicted) && (ActivityPatterns[i][j].mode == PATTERN_MODE_ACTIVE))
				{
					predicted = true;
					ActivityStatistics[i].truePositives++;
					errorTime = (ActivityPatterns[i][j].time + ActivityPatterns[i][j].mean - activityTime);
					ActivityStatistics[i].error += abs (errorTime);
				}
				ActivityPatterns[i][j].mode = PATTERN_MODE_INACTIVE;
			}
			if (activity == ActivityPatterns[i][j].relativeActivity)
			{
				// If we see the pattern's relative activity again without seeing its predicted activity, then false positive
				// and reset its time
				if (ActivityPatterns[i][j].mode == PATTERN_MODE_ACTIVE)
				{
					ActivityStatistics[i].falsePositives++;
					ActivityPatterns[i][j].time = activityTime;
				}
				if (ActivityPatterns[i][j].mode == PATTERN_MODE_INACTIVE)
				{
					ActivityPatterns[i][j].mode = PATTERN_MODE_ACTIVE;
					ActivityPatterns[i][j].time = activityTime;
				}
			}
		}
	}
	// if the activity was not predicted, then false negative
	if (! predicted)
	{
		ActivityStatistics[activity].falseNegatives++;
	}
}
